from typing import Optional, Union

from numpy import ndarray

from mygrad.tensor_base import Tensor
from mygrad.typing import ArrayLike, DTypeLikeReals, Mask
from mygrad.ufuncs._ufunc_creators import ufunc_creator

from .ops import (
    Add,
    AddSequence,
    Divide,
    Multiply,
    MultiplySequence,
    Negative,
    Positive,
    Power,
    Reciprocal,
    Square,
    Subtract,
)

__all__ = [
    "add",
    "add_sequence",
    "divide",
    "multiply",
    "multiply_sequence",
    "negative",
    "positive",
    "power",
    "reciprocal",
    "square",
    "subtract",
    "true_divide",
]


@ufunc_creator(Add)
def add(
    x1: ArrayLike,
    x2: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Add the arguments element-wise.

    This docstring was adapted from that of numpy.add [1]_

    Parameters
    ----------
    x1, x2 : ArrayLike
        The arrays to be added.
        If ``x1.shape != x2.shape``, they must be broadcastable to a common
        shape (which becomes the shape of the output). Non-tensor array-likes are
        treated as constants.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    out : Optional[Union[ndarray, Tensor]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    Returns
    -------
    add : Tensor
        The sum of `x1` and `x2`, element-wise.

    Notes
    -----
    Equivalent to `x1` + `x2` in terms of tensor broadcasting.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.add.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.add(1.0, 4.0)
    Tensor(5.0)
    >>> x1 = mg.tensor([[0., 1., 2.],
    ...                 [3., 4., 5.],
    ...                 [6., 7., 8.]])
    >>> x2 = mg.tensor([0., 1., 2.])
    >>> mg.add(x1, x2)
    Tensor([[  0.,   2.,   4.],
            [  3.,   5.,   7.],
            [  6.,   8.,  10.]])
    """
    ...


@ufunc_creator(Subtract)
def subtract(
    x1: ArrayLike,
    x2: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Subtract the arguments element-wise.

    This docstring was adapted from that of numpy.subtract [1]_

    Parameters
    ----------
    x1, x2 : ArrayLike
        The arrays to be subtracted from each other.
        If ``x1.shape != x2.shape``, they must be broadcastable to a common
        shape (which becomes the shape of the output). Non-tensor array-likes are
        treated as constants.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    out : Optional[Union[ndarray, Tensor]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    Returns
    -------
    subtract : Tensor
        The difference of `x1` and `x2`, element-wise.

    Notes
    -----
    Equivalent to ``x1 - x2`` in terms of tensor broadcasting.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.subtract.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.subtract(1.0, 4.0)
    Tensor(-3.0)

    >>> x1 = mg.tensor([[0., 1., 2.],
    ...                 [3., 4., 5.],
    ...                 [6., 7., 8.]])
    >>> x2 = mg.tensor([0., 1., 2.])
    >>> mg.subtract(x1, x2)
    Tensor([[ 0.,  0.,  0.],
            [ 3.,  3.,  3.],
            [ 6.,  6.,  6.]])
    """
    ...


@ufunc_creator(Multiply)
def multiply(
    x1: ArrayLike,
    x2: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Multiply the arguments element-wise.

    This docstring was adapted from that of numpy.multiply [1]_

    Parameters
    ----------
    x1, x2 : ArrayLike
        Input arrays to be multiplied.
        If ``x1.shape != x2.shape``, they must be broadcastable to a common
        shape (which becomes the shape of the output). Non-tensor array-likes
        are treated as constants.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    out : Optional[Union[ndarray, Tensor]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    Returns
    -------
    multiply : Tensor
        The product of `x1` and `x2`, element-wise.

    Notes
    -----
    Equivalent to `x1` * `x2` in terms of tensor broadcasting.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.multiply.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.multiply(2.0, 4.0)
    Tensor(8.0)

    >>> x1 = mg.tensor([[0., 1., 2.],
    ...                 [3., 4., 5.],
    ...                 [6., 7., 8.]])
    >>> x2 = mg.tensor([0., 1., 2.])
    >>> mg.multiply(x1, x2)
    Tensor([[  0.,   1.,   4.],
            [  0.,   4.,  10.],
            [  0.,   7.,  16.]])
    """
    ...


@ufunc_creator(Divide)
def true_divide(
    x1: ArrayLike,
    x2: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Divide the arguments element-wise.

    This docstring was adapted from that of numpy.true_divide [1]_

    Parameters
    ----------
    x1 : ArrayLike
        Dividend array.

    x2 : ArrayLike
        Divisor array.
        If ``x1.shape != x2.shape``, they must be broadcastable to a common
        shape (which becomes the shape of the output). Non-tensor array-likes
        are treated as constants.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    out : Optional[Union[ndarray, Tensor]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    Returns
    -------
    true_divide : Tensor
        The quotient of `x1` with `x2`, element-wise.

    Notes
    -----
    In Python, ``//`` is the floor division operator and ``/`` the
    true division operator.  The ``true_divide(x1, x2)`` function is
    equivalent to true division in Python.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.true_divide.html

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.arange(5)
    >>> mg.true_divide(x, 4)
    Tensor([ 0.  ,  0.25,  0.5 ,  0.75,  1.  ])

    >>> x/4
    Tensor([ 0.  ,  0.25,  0.5 ,  0.75,  1.  ])

    >>> x // 4
    Tensor([0, 0, 0, 0, 1], dtype=int32)

    Floor division with a tensor always produces a constant

    >>> (x // 4).constant
    True
    """
    ...


divide = true_divide


@ufunc_creator(Power)
def power(
    x1: ArrayLike,
    x2: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """First tensor elements raised to powers from second tensor, element-wise.

    Raise each base in `x1` to the positionally-corresponding power in
    `x2`.  `x1` and `x2` must be broadcastable to the same shape. Note that an
    integer type raised to a negative integer power will raise a ValueError.

    This docstring was adapted from that of numpy.power [1]_

    Parameters
    ----------
    x1 : ArrayLike
        The bases.

    x2 : ArrayLike
        The exponents.
        If ``x1.shape != x2.shape``, they must be broadcastable to a common
        shape (which becomes the shape of the output). Non-tensor array-likes
        are treated as constants.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    out : Optional[Union[ndarray, Tensor]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    Returns
    -------
    power : Tensor
        The combination of `x1` and `x2`, element-wise.

    See Also
    --------
    float_power : power function that promotes integers to float

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.power.html

    Examples
    --------
    Cube each element in a list.

    >>> import mygrad as mg
    >>> x1 = range(6)
    >>> x1
    [0, 1, 2, 3, 4, 5]
    >>> mg.power(x1, 3)
    Tensor([  0,   1,   8,  27,  64, 125])

    Raise the bases to different exponents.

    >>> x2 = [1.0, 2.0, 3.0, 3.0, 2.0, 1.0]
    >>> mg.power(x1, x2)
    Tensor([  0.,   1.,   8.,  27.,  16.,   5.])

    The effect of broadcasting.

    >>> x2 = mg.tensor([[1, 2, 3, 3, 2, 1], [1, 2, 3, 3, 2, 1]])
    >>> x2
    Tensor([[1, 2, 3, 3, 2, 1],
            [1, 2, 3, 3, 2, 1]])
    >>> mg.power(x1, x2)
    Tensor([[ 0,  1,  8, 27, 16,  5],
            [ 0,  1,  8, 27, 16,  5]])
    """
    ...


@ufunc_creator(Negative)
def negative(
    x: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Negates the tensor element-wise.

    This docstring was adapted from that of numpy.negative [1]_

    Parameters
    ----------
    x : ArrayLike or scalar
        Input tensor.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    Returns
    -------
    negative : Tensor
        The combination of `x1` and `x2`, element-wise.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.negative.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.negative([1.,-1.])
    Tensor([-1.,  1.])
    """
    ...


@ufunc_creator(Positive)
def positive(
    x: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Returns a copy of the tensor.

    This docstring was adapted from that of numpy.positive [1]_

    Parameters
    ----------
    x : ArrayLike
        Input array.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    Returns
    -------
    positive : Tensor

    Notes
    -----
    Equivalent to `x.copy()`, but only defined for types that support
    arithmetic.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.positive.html
    """
    ...


@ufunc_creator(Reciprocal)
def reciprocal(
    x: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Return the reciprocal of the argument element-wise.

    This docstring was adapted from that of numpy.reciprocal [1]_

    Parameters
    ----------
    x : ArrayLike
        Input array.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    Returns
    -------
    reciprocal : Tensor

    Notes
    -----
    .. note::
        This function is not designed to work with integers.

    For integer arguments with absolute value larger than 1 the result is
    always zero because of the way Python handles integer division.  For
    integer zero the result is an overflow.

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.reciprocal.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.reciprocal(2.)
    Tensor(0.5)
    >>> mg.reciprocal([1, 2., 3.33])
    Tensor([ 1.       ,  0.5      ,  0.3003003])
    """
    ...


@ufunc_creator(Square)
def square(
    x: ArrayLike,
    out: Optional[Union[ndarray, Tensor]] = None,
    *,
    where: Mask = True,
    dtype: DTypeLikeReals = None,
    constant: Optional[bool] = None,
) -> Tensor:  # pragma: no cover
    """Return the square of the argument element-wise.

    This docstring was adapted from that of numpy.square [1]_

    Parameters
    ----------
    x : ArrayLike
        Input data.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    dtype : Optional[DTypeLikeReals]
        The dtype of the resulting tensor.

    out : Optional[Union[Tensor, ndarray]]
        A location into which the result is stored. If provided, it must have
        a shape that the inputs broadcast to. If not provided or None,
        a freshly-allocated tensor is returned.

    where : Mask
        This condition is broadcast over the input. At locations where the
        condition is True, the ``out`` tensor will be set to the ufunc result.
        Elsewhere, the ``out`` tensor will retain its original value.
        Note that if an uninitialized `out` tensor is created via the default
        ``out=None``, locations within it where the condition is False will
        remain uninitialized.

    Returns
    -------
    square : Tensor

    See Also
    --------
    sqrt
    power

    References
    ----------
    .. [1] Retrieved from https://numpy.org/doc/stable/reference/generated/numpy.square.html

    Examples
    --------
    >>> import mygrad as mg
    >>> mg.square([100., 1000.])
    array([10.,  100.])
    """
    ...


def multiply_sequence(*variables: ArrayLike, constant: Optional[bool] = None) -> Tensor:
    """``f(a, b, ...) -> a * b * ...``

    Multiply a sequence of tensors.

    Parameters
    ----------
    variables : ArrayLike
        A sequence of broadcast-compatible tensors. Non-tensor array-likes are
        treated as constants.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor

    Notes
    -----
    It is more efficient to back-propagate through this
    function than it is through a computational graph
    with N-1 corresponding multiplication operations.

    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.tensor([1. , 2.])
    >>> y = mg.tensor([-1.])
    >>> z = mg.tensor([[1.]])
    >>> out = mg.multiply_sequence(x, y, z); out
        Tensor([[-1., -2.]])

    >>> out.backward()
    >>> x.grad
    array([-1., -1.])
    >>> y.grad
    array([3.])
    >>> z.grad
    array([[-3.]])
    """
    if len(variables) < 2:
        raise ValueError(
            f"`multiply_sequence` requires at least two inputs, got {len(variables)} inputs"
        )
    return Tensor._op(MultiplySequence, *variables, constant=constant)


def add_sequence(*variables: ArrayLike, constant: Optional[bool] = None) -> Tensor:
    """``f(a, b, ...) -> a + b + ...``

    Add a sequence of tensors.

    Parameters
    ----------
    variables : ArrayLike
        A sequence of broadcast-compatible tensors. Non-tensor array-likes are
        treated as constants.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor

    Notes
    -----
    It is more efficient to back-propagate through this
    function than it is through a computational graph
    with N-1 corresponding addition operations.


    Examples
    --------
    >>> import mygrad as mg
    >>> x = mg.tensor([1. , 2.])
    >>> y = mg.tensor([-1.])
    >>> z = mg.tensor([[1.]])
    >>> out = mg.add_sequence(x, y, z); out
        Tensor([[1., 2.]])

    >>> out.backward()
    >>> x.grad
    array([1., 1.])
    >>> y.grad
    array([2.])
    >>> z.grad
    array([[2.]])
    """
    if len(variables) < 2:
        raise ValueError(
            f"`add_sequence` requires at least two inputs, got {len(variables)} inputs"
        )
    return Tensor._op(AddSequence, *variables, constant=constant)



from typing import Optional

import numpy as np

import mygrad._utils.graph_tracking as _tracking
from mygrad.operation_base import Operation
from mygrad.tensor_base import Tensor
from mygrad.typing import ArrayLike

from ._utils import check_loss_inputs


class MulticlassHinge(Operation):
    def __call__(self, a, y, hinge=1.0):
        """Computes the average multiclass hinge loss.

        Parameters
        ----------
        a : mygrad.Tensor, shape=(N, C)
            The C class scores for each of the N pieces of data.

        y : numpy.ndarray, shape=(N,)
            The correct class-index, in [0, C), for each datum.

        Returns
        -------
        The average multiclass hinge loss

        Raises
        ------
        TypeError
            `y_true` must be an integer-type array-like object

        ValueError
            `x` must be a 2-dimensional array-like object
            `y_true` must be a shape-(N,) array-like object"""

        check_loss_inputs(a, y)
        self.variables = (a,)
        scores = a.data
        correct_labels = (range(len(y)), y)
        correct_class_scores = scores[correct_labels]  # Nx1

        M = scores - correct_class_scores[:, np.newaxis] + hinge  # NxC margins
        not_thresh = np.where(M <= 0)
        Lij = M
        Lij[not_thresh] = 0
        Lij[correct_labels] = 0
        if _tracking.TRACK_GRAPH:
            TMP = np.ones(M.shape, dtype=float)
            TMP[not_thresh] = 0
            TMP[correct_labels] = 0  # NxC; 1 where margin > 0
            TMP[correct_labels] = -1 * TMP.sum(axis=-1)
            self.back = TMP
            self.back /= scores.shape[0]
        return np.sum(Lij) / scores.shape[0]

    def backward_var(self, grad, index, **kwargs):
        return grad * self.back


def multiclass_hinge(
    x: ArrayLike,
    y_true: ArrayLike,
    hinge: float = 1.0,
    *,
    constant: Optional[bool] = None
) -> Tensor:
    """Computes the average multiclass hinge loss.

    Parameters
    ----------
    x : ArrayLike, shape=(N, K)
        The K class scores for each of the N pieces of data.

    y_true : ArrayLike, shape=(N,)
        The correct class-indices, in [0, K), for each datum.

    hinge : float
        The size of the "hinge" outside of which a nonzero loss
        is incurred.

    constant : bool, optional(default=False)
        If ``True``, the returned tensor is a constant (it
        does not back-propagate a gradient)

    Returns
    -------
    Tensor, shape-() (scalar)
        The average multiclass hinge loss

    Raises
    ------
    TypeError
        `y_true` must be an integer-type array-like object

    ValueError
        `x` must be a 2-dimensional array-like object
        `y_true` must be a shape-(N,) array-like object
    """
    return Tensor._op(MulticlassHinge, x, op_args=(y_true, hinge), constant=constant)



from typing import Optional

import numpy as np

from mygrad import Tensor
from mygrad.typing import Shape


def rand(*shape: int, constant: Optional[bool] = None) -> Tensor:
    """Create a Tensor of the given shape and populate it with random
    samples from a uniform distribution over [0, 1).

    Parameters
    ----------
    shape: d0, d1, ... dn : int, optional
        The dimensions of the returned array, must be non-negative.
        If no argument is given a single Python float is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor
        A ``shape``--shaped Tensor of floating-point samples from the uniform distribution
        over [0, 1), or a single such float if no parameters were supplied.

    Examples
    --------
    >>> from mygrad.random import rand
    >>> rand(3,4)
    Tensor([[0.9805903 , 0.82640985, 0.88230632, 0.73099815],
            [0.24845968, 0.12532893, 0.63171607, 0.32543228],
            [0.66029533, 0.79285341, 0.54967228, 0.25178508]])
    """

    return Tensor(np.random.rand(*shape), constant=constant, copy=False)


def randint(low, high=None, shape: Optional[Shape] = None, dtype=int) -> Tensor:
    """Return random integers from the “discrete uniform” distribution of the specified dtype in the
    “half-open” interval [low, high).

    If high is None (the default), then results are from [0, low).

    Parameters
    ----------
    low: int or array-like of ints
        Lowest (signed) integers to be drawn from the distribution
        (unless high=None, in which case this parameter is one above the highest such integer).

    high: int or array-like of ints, optional
        If provided, one above the largest (signed) integer to be drawn from the
        distribution (see above for behavior if high=None). If array-like, must contain integer values

    shape: int or tuple of ints, optional
        Output shape. If the given shape is, e.g., (m, n, k), then m * n * k samples are drawn.
        Default is None, in which case a single value is returned.

    dtype: dtype, optional
        Desired dtype of the result. Byteorder must be native. The default value is int.

    Returns
    -------
    int or mygrad.Tensor of ints
        ``shape``-shaped array of random integers from the appropriate distribution,
        or a single such random int if size not provided.

    Examples
    --------
    >>> from mygrad.random import randint
    >>> randint(low=1, high=7, shape=(2,5))
    Tensor([[2, 4, 1, 5, 1],
            [6, 2, 5, 4, 6]])

    >>> randint(low=4, high=100)
    Tensor(57)
    """

    return Tensor(np.random.randint(low, high, shape, dtype), copy=False)


def randn(*shape: int, constant: Optional[bool] = None) -> Tensor:
    """Return a sample (or samples) from the “standard normal” distribution.

    Parameters
    ----------
    shape: shape: d0, d1, ... dn : int, optional
        The dimensions of the returned array, must be non-negative.
        If no argument is given a single Python float is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.

    Returns
    -------
    mygrad.Tensor
        A ``shape``-shaped Tensor of floating-point samples from the standard normal distribution,
        or a single such float if no parameters were supplied.

    Examples
    --------
    >>> from mygrad.random import randn
    >>> randn(3, 3, 2)
    Tensor([[[-0.45664135,  0.05060159],
             [ 1.36883177, -0.46084292],
             [-0.76647664,  0.81667174]],

            [[ 0.08336453, -1.35104408],
             [ 0.73187355,  1.33405382],
             [ 0.28411209, -0.18047323]],

            [[-0.2239412 , -0.09170368],
             [-0.39175898,  0.81260396],
             [-1.28788909, -1.52525778]]])
    """

    return Tensor(np.random.randn(*shape), constant=constant, copy=False)


def random(shape: Optional[Shape] = None, *, constant: Optional[bool] = None) -> Tensor:
    """Return random floats in the half-open interval [0.0, 1.0).

    To create a random sample of a given shape on the interval [a, b), call
    (b-a) * random(shape) + a

    Parameters
    ----------
    shape: int or tuple of ints, optional
        Output shape. If the given shape is, e.g., (m, n, k), then m * n * k samples are drawn.
        Default is None, in which case a single value is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.
    Returns
    -------
    int or mygrad.Tensor of ints
        ``shape``-shaped array of random integers from the appropriate distribution, or a
        single such random int if size not provided.

    Examples
    --------
    >>> from mygrad.random import random
    >>> random((2, 4))
    Tensor([[0.14928578, 0.28812813, 0.56885892, 0.49555962],
            [0.19780163, 0.51162365, 0.7849505 , 0.47864586]])
    """

    return Tensor(np.random.random(shape), constant=constant, copy=False)


def random_sample(
    shape: Optional[Shape] = None, *, constant: Optional[bool] = None
) -> Tensor:
    """Return random floats in the half-open interval [0.0, 1.0).

    Results are from the “continuous uniform” distribution over the stated interval.

    To create a random sample of a given shape on the interval [a, b), call
    (b-a) * random_sample(shape) + a

    Parameters
    ----------
    shape: int or tuple of ints, optional
        Output shape. If the given shape is, e.g., (m, n, k), then m * n * k samples are drawn.
        Default is None, in which case a single value is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.
    Returns
    -------
    int or mygrad.Tensor of ints
        ``shape``-shaped array of random integers from the appropriate distribution,
        or a single such random int if size not provided.

    Examples
    --------
    >>> from mygrad.random import random_sample
    >>> random_sample((3, 2))
    Tensor([[0.76490814, 0.69378441],
            [0.65228375, 0.68395309],
            [0.08228869, 0.03191064]])

    >>> random_sample()
    Tensor(0.47644928)
    """

    return Tensor(np.random.random_sample(shape), constant=constant, copy=False)


def ranf(shape: Optional[Shape] = None, *, constant: Optional[bool] = None) -> Tensor:
    """Return random floats in the half-open interval [0.0, 1.0).

    To create a random sample of a given shape on the interval [a, b), call
    (b-a) * ranf(shape) + a

    Parameters
    ----------
    shape: int or tuple of ints, optional
        Output shape. If the given shape is, e.g., (m, n, k), then m * n * k samples are drawn.
        Default is None, in which case a single value is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.
    Returns
    -------
    int or mygrad.Tensor of ints
        ``shape``-shaped array of random integers from the appropriate distribution, or
        a single such random int if size not provided.

    Examples
    --------
    >>> from mygrad.random import ranf
    >>> ranf((2, 3, 1))
    Tensor([[[0.9343681 ],
             [0.29573802],
             [0.84759669]],

            [[0.34563731],
             [0.68601617],
             [0.02388943]]])

    >>> ranf()
    Tensor(0.77739196)
    """

    return Tensor(np.random.ranf(shape), constant=constant, copy=False)


def sample(shape: Optional[Shape] = None, *, constant: Optional[bool] = None) -> Tensor:
    """Return random floats in the half-open interval [0.0, 1.0).

    To create a random sample of a given shape on the interval [a, b), call
    (b-a) * sample(shape) + a

    Parameters
    ----------
    shape: int or tuple of ints, optional
        Output shape. If the given shape is, e.g., (m, n, k), then m * n * k samples are drawn.
        Default is None, in which case a single value is returned.

    constant : Optional[bool]
        If ``True``, this tensor is treated as a constant, and thus does not
        facilitate back propagation (i.e. ``constant.grad`` will always return
        ``None``).

        Defaults to ``False`` for float-type data.
        Defaults to ``True`` for integer-type data.

        Integer-type tensors must be constant.
    Returns
    -------
    int or mygrad.Tensor of ints
        ``shape``-shaped array of random integers from the appropriate distribution,
        or a single such random int if size not provided.

    Examples
    --------
    >>> from mygrad.random import sample
    >>> sample((3, 4))
    Tensor([[0.47263933, 0.10928814, 0.19737707, 0.30879006],
            [0.49870689, 0.05849937, 0.21095352, 0.09778017],
            [0.405788  , 0.91888808, 0.15061143, 0.63140668]])

    >>> sample()
    Tensor(0.50690423)
    """

    return Tensor(np.random.sample(shape), constant=constant, copy=False)


def seed(seed_number):
    """Seed the generator.

    Simply used NumPy's random state - i.e. this is equivalent to ``numpy.random.seed``.

    Parameters
    ----------
    seed_number : int or 1-d array_like, optional
        Seed for RandomState. Must be convertible to 32 bit unsigned integers.

    Examples
    --------
    >>> from mygrad.random import seed, random
    >>> seed(0)
    >>> random((2, 4))
    Tensor([[0.5488135 , 0.71518937, 0.60276338, 0.54488318],
            [0.4236548 , 0.64589411, 0.43758721, 0.891773  ]])

    >>> seed(1)
    >>> random((2, 4))
    Tensor([[4.17022005e-01, 7.20324493e-01, 1.14374817e-04, 3.02332573e-01],
            [1.46755891e-01, 9.23385948e-02, 1.86260211e-01, 3.45560727e-01]]

    >>> seed(0)
    >>> random((2,4))
    Tensor([[0.5488135 , 0.71518937, 0.60276338, 0.54488318],
            [0.4236548 , 0.64589411, 0.43758721, 0.891773  ]])
    """

    np.random.seed(seed_number)



from typing import Optional, Tuple, Union

import numpy as np

from mygrad.tensor_base import Tensor, implements_numpy_override
from mygrad.typing import ArrayLike

__all__ = ["any", "argmin", "argmax"]


@implements_numpy_override()
def argmax(
    a: ArrayLike, axis: Optional[int] = None, out: Optional[np.ndarray] = None
) -> np.ndarray:
    """Returns the indices of the maximum values along an axis.

    Parameters
    ----------
    a: array_like

    axis: int, optional
        By default, the index is into the flattened array, otherwise along the specified axis.

    out: numpy.array, optional
        If provided, the result will be inserted into this array. It should be of the appropriate shape and dtype.

    Returns
    -------
    numpy.ndarray[int]"""

    a = a.data if isinstance(a, Tensor) else a
    return np.argmax(a, axis, out)


@implements_numpy_override()
def argmin(
    a: ArrayLike, axis: Optional[int] = None, out: Optional[np.ndarray] = None
) -> np.ndarray:
    """Returns the indices of the minimum values along an axis.

    Parameters
    ----------
    a: array_like

    axis: int, optional
        By default, the index is into the flattened array, otherwise along the specified axis.

    out: numpy.array, optional
        If provided, the result will be inserted into this array. It should be of the appropriate shape and dtype.

    Returns
    -------
    numpy.ndarray[int]"""

    a = a.data if isinstance(a, Tensor) else a
    return np.argmin(a, axis, out)


@implements_numpy_override()
def any(
    a: ArrayLike,
    axis: Optional[Union[int, Tuple[int, ...]]] = None,
    out: Optional[np.ndarray] = None,
    keepdims: bool = False,
) -> np.ndarray:
    """Test whether any array or Tensor element along a given axis evaluates to True.

    Returns single boolean if `axis` is ``None``

    This documentation was adapted from ``numpy.add``

    Parameters
    ----------
    a : array_like

    axis : None or int or tuple of ints, optional
        Axis or axes along which a logical OR reduction is performed.
        The default (``axis=None``) is to perform a logical OR over all
        the dimensions of the input array. `axis` may be negative, in
        which case it counts from the last to the first axis.
        If this is a tuple of ints, a reduction is performed on multiple
        axes, instead of a single axis or all the axes as before.

    out : ndarray, optional
        Alternate output array in which to place the result.  It must have
        the same shape as the expected output and its type is preserved
        (e.g., if it is of type float, then it will remain so, returning
        1.0 for True and 0.0 for False, regardless of the type of `a`).
        See `ufuncs-output-type` for more details.

    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left
        in the result as dimensions with size one. With this option,
        the result will broadcast correctly against the input array.
        If the default value is passed, then `keepdims` will not be
        passed through to the `any` method of sub-classes of
        `ndarray`, however any non-default value will be.  If the
        sub-class' method does not implement `keepdims` any
        exceptions will be raised.

    Returns
    -------
    any : bool or ndarray
        A new boolean or `ndarray` is returned unless `out` is specified,
        in which case a reference to `out` is returned.

    See Also
    --------
    Tensor.any : equivalent method

    """

    a = a.data if isinstance(a, Tensor) else a
    return np.any(a, axis=axis, out=out, keepdims=keepdims)



"""
Provides utilities responsible for locking/releasing array writeability.
"""
import os
from collections import Counter, defaultdict
from typing import TYPE_CHECKING, DefaultDict, Dict, Generator, Iterable
from weakref import finalize, ref

import numpy as np

from mygrad._utils import ContextTracker, WeakRefIterable

if TYPE_CHECKING:  # pragma: no cover
    from typing import Set

    from mygrad import Tensor
    from mygrad._utils import WeakRef

# arr-id -> num active ops involving arr
_array_counter = Counter()  # type: Counter[int, int]

# arr-id -> weak-ref of arr, for arrays participating in live ops
_array_tracker = dict()  # type: Dict[int, WeakRef[np.ndarray]]

# maps base-array ID to ID of view that can't be unlocked until
# base is unlocked
_views_waiting_for_unlock = defaultdict(
    set
)  # type: DefaultDict[int, Set[int]] # base-id -> set of view-ids

__all__ = [
    "lock_arr_writeability",
    "release_writeability_lock_on_op",
    "mem_guard_off",
    "mem_guard_on",
    "mem_guard_active",
]


def array_is_tracked(arr: np.ndarray) -> bool:
    """Returns True if the provided array, or a view of it, is currently
    involved in one or more mygrad operation."""
    arr_id = id(arr)
    return arr_id in _array_tracker and _array_tracker[arr_id]() is not None


def lock_arr_writeability(arr: np.ndarray, force_lock: bool = False) -> np.ndarray:
    """Increments the count of active ops that an array is involved in
    and makes the array read-only

    Parameters
    ----------
    arr : numpy.ndarray

    force_lock : bool, optional (default=False)
        If True, and array that is already read-only will be tracked
        for unlocking

    Returns
    -------
    numpy.ndarray
        The locked array"""
    arr_id = id(arr)
    if not array_is_tracked(arr):
        if (
            not force_lock
            and not arr.flags.writeable
            and (arr.base is None or not array_is_tracked(arr.base))
        ):
            # array is natively read-only; don't do anything
            return arr
        # keeps track of array so we can clean up the array
        # counter when tracked arrays fall out of scope
        _array_tracker[arr_id] = ref(arr)
        _array_counter[arr_id] = 1
    else:
        _array_counter[arr_id] += 1
    if arr.flags.writeable is True:
        arr.flags.writeable = False
    return arr


def unique_arrs_and_bases(
    tensors: Iterable["Tensor"],
) -> Generator[np.ndarray, None, None]:
    """
    Yields unique (by-ID) arrays from an iterable. If an array
    has a base, the base is yielded first (assuming that base
    object has not already been yielded).
    """
    seen = set()
    for t in tensors:
        arr = t.data
        arr_id = id(arr)
        if arr_id not in seen:

            # important note!
            # We must yield array bases first so that base's
            # writeability is restored first.
            # Then view's writeability can be restored
            if arr.base is not None:
                base_id = id(arr.base)
                if base_id not in seen:
                    seen.add(base_id)
                    yield arr.base
            seen.add(arr_id)
            yield arr


def _release_lock_on_arr_writeability(arr: np.ndarray):
    """
    Decrements the number of active ops the array participates in.
    An array no longer participating in any ops will have its
    writeability restored.
    """
    arr_id = id(arr)
    num_active_ops = _array_counter[arr_id]

    if num_active_ops == 1:
        # final active op involving array is being de-referenced:
        # okay to unlock array
        del _array_counter[arr_id]

        if arr.base is not None and arr.base.flags.writeable is False:
            # Array is view and must wait until its base is released
            # before it can be unlocked
            # Thus we are still tracking this array
            _views_waiting_for_unlock[id(arr.base)].add(arr_id)
        else:
            # we no longer need to track the array
            arr.flags.writeable = True
            _array_tracker.pop(arr_id, None)
            if not _array_tracker and _views_waiting_for_unlock:
                # If no arrays are being tracked, then there can't
                # be any views waiting to be unlocked.
                # Clean up!
                _views_waiting_for_unlock.clear()
    elif num_active_ops > 0:
        _array_counter[arr_id] = num_active_ops - 1

    if (
        arr.base is None
        and arr.flags.writeable
        and (arr_id in _views_waiting_for_unlock)
    ):
        # array was base of view waiting to be unlocked..
        #
        # Either:
        #    view no longer exists
        #    or view is involved in new op
        #    or view can now get unlocked
        # under all conditions view will no longer be waiting to be unlocked
        for view_arr_id in tuple(_views_waiting_for_unlock[arr_id]):

            if _array_counter[view_arr_id] > 0:
                # view involved in new op
                continue

            _views_waiting_for_unlock[arr_id].remove(view_arr_id)

            try:
                view_arr = _array_tracker.pop(view_arr_id)()
                if view_arr is None:
                    continue
            except KeyError:
                # view array is no longer available for unlocking
                continue

            try:
                view_arr.flags.writeable = True
            except ValueError:  # pragma: no cover
                # sometimes this raises.. but it is not
                # reproducible and is very rare
                pass

        if not _views_waiting_for_unlock[arr_id]:
            _views_waiting_for_unlock.pop(arr_id)


def release_writeability_lock_on_op(arr_refs: WeakRefIterable[np.ndarray]):
    """Marks each array (and for a view, its base) to have its
    writeability lock released.

    An array is made writeable only once all of its locks
    have been released.

    Parameters
    ----------
    arr_refs : WeakRefIterable[np.ndarray]
        The arrays to be unlocked. Only one lock is released
        on each array, even if the same array occurs
        multiple times in the iterable."""
    for arr in arr_refs:
        _release_lock_on_arr_writeability(arr)


MEM_GUARD = os.environ.get("MYGRAD_MEM_GUARD", True)

if MEM_GUARD in {"True", "true", "1", 1, True}:
    MEM_GUARD = True
elif MEM_GUARD in {"False", "false", "0", 0, False}:  # pragma: no cover
    MEM_GUARD = False
else:  # pragma: no cover
    from warnings import warn

    warn(
        f"Environment variable MYGRAD_MEM_GUARD was set to an unknown value {MEM_GUARD}. "
        f"Proceeding with `MEM_GUARD=True`"
    )
    MEM_GUARD = True


class MemStateContext(ContextTracker):
    @property
    def state(self):
        return MEM_GUARD

    @state.setter
    def state(self, value: bool):
        if not isinstance(value, bool):  # pragma: no cover
            raise TypeError(
                f"MEM_GUARD must be set to a boolean value, got {value} (type={type(value)})"
            )

        global MEM_GUARD
        MEM_GUARD = value


class _NoMemGuard(MemStateContext):
    """A context manager used to suspend memory-locking behavior

    Examples
    --------
    >>> from mygrad import  mem_guard_off
    >>> with mem_guard_off:
    ...     # array-memory locking is turned off
    ...     pass
    ... # previous memory-locking behavior is restored

    This can also be used as a decorator

    >>> @mem_guard_off
    >>> def f():
    ...     # array-memory locking is turned off within function
    ...     return

    """

    _enter_set_value = False


class _WithMemGuard(MemStateContext):
    """A context manager used to enable memory-locking behavior

    Examples
    --------
    >>> from mygrad import mem_guard_on
    >>> with mem_guard_on:
    ...     # array-memory locking is turned on
    ...     pass
    ... # previous memory-locking behavior is restored

    This can also be used as a decorator

    >>> @mem_guard_on
    >>> def f():
    ...     # array-memory locking is turned on within function
    ...     return

    """

    _enter_set_value = True


mem_guard_off = _NoMemGuard()
mem_guard_on = _WithMemGuard()


def turn_memory_guarding_off():
    """Globally disables all memory-guarding mechanisms, except
    for in contexts where they are explicitly enabled.

    Notes
    -----
    With memory guarding disabled, arrays participating in active
    computational graphs are not protected from being mutated by
    the user. Mutating such an array will corrupt the derivatives
    that are computed via back-propagation, and will produce
    incorrect results.

    This can speed up computations involving many small tensors
    substantially.

    If you want to disable memory guarding at the system level, you
    can set the system environment variable MYGRAD_MEM_GUARD=False.
    NOTE THAT THIS IS NOT RECOMMENDED.

    See Also
    --------
    turn_memory_guarding_on : Globally enables all memory-guarding mechanisms
    mem_guard_off : context manager & decorator for suspending memory guarding
    mem_guard_on : context manager & decorator for enabling memory guarding

    Examples
    --------
    The following demonstrates how one can unwittingly corrupt
    backpropagation through a computational graph

    >>> import mygrad as mg
    >>> import numpy as np
    >>> mg.turn_memory_guarding_off()  # speeds up calculations, but with risks involved..
    >>> x = np.arange(3.)
    >>> y = mg.ones_like(x)
    >>> z = x * y
    >>> x[:] = 0  # mutates x, corrupting state associated with z
    >>> z.backward()
    >>> y.grad  # would be array([0., 1., 2.]) if graph wasn't corrupted
    array([0., 0., 0.])
    """
    global MEM_GUARD
    MEM_GUARD = False


def turn_memory_guarding_on():
    """Globally enables all memory-guarding mechanisms, except
    for in contexts where they are explicitly disabled.

    Notes
    -----
    Memory guarding is enabled by default. It ensures that arrays
    that are participating in computational graphs cannot be mutated
    (at least unwittingly..), which provides important assurances that
    the state of the computational graph is not corrupted for
    back-propagation.

    Memory guarding can slow down computations involving many small tensors.
    Realistic worst-case benchmarks suggest a ~50% slowdown.

    If performance is important, it is recommended that you test your code leaving
    memory guarding enabled. Presuming the code runs without any errors regarding
    writing to read-only arrays, you can proceed to disable memory guarding and
    enjoy the concomitant speedups.

    Note also that running your code in a `no_autodiff` context will automatically
    disable memory guarding.

    See Also
    --------
    turn_memory_guarding_off : Globally enables all memory-guarding mechanisms
    mem_guard_off : context manager & decorator for suspending memory guarding
    mem_guard_on : context manager & decorator for enabling memory guarding
    no_autodiff : context manager for disabling graph-tracking for back propagation

    Examples
    --------
    The following demonstrates how memory guarding prevents one from
    unwittingly corrupting an active computational graph

    >>> import mygrad as mg
    >>> import numpy as np
    >>> # (Note that memory guarding is on by default, so
    >>> # this call isn't usually needed...)
    >>> mg.turn_memory_guarding_on()
    >>> x = np.arange(3.)
    >>> y = mg.ones_like(x)
    >>> z = x * y
    >>> try:
    ...     x[:] = 0  # raises because `x` is made read-only
    ... except ValueError:
    ...     pass
    >>> z.backward()
    >>> y.grad  # correct gradient is computed
    array([0., 1., 2.])
    """
    global MEM_GUARD
    MEM_GUARD = True


def mem_guard_active() -> bool:
    """Indicates whether or not memory guarding is active.

    See Also
    --------
    turn_memory_guarding_on : Globally enables all memory-guarding mechanisms
    turn_memory_guarding_off : Globally enables all memory-guarding mechanisms
    mem_guard_off : context manager & decorator for suspending memory guarding
    mem_guard_on : context manager & decorator for enabling memory guarding
    """
    return MEM_GUARD


def force_lock_tensor_and_creators(tensor: "Tensor"):
    unique_arrs = tuple(
        lock_arr_writeability(arr)
        for arr in unique_arrs_and_bases(tensor.creator.variables)
    )
    lock_arr_writeability(tensor.data, force_lock=True)
    tensor_refs = WeakRefIterable(unique_arrs)
    tensor_refs.append(tensor.data)
    finalize(
        tensor.creator,
        release_writeability_lock_on_op,
        tensor_refs,
    )



from .batchnorm import batchnorm
from .conv import conv_nd
from .pooling import max_pool

__all__ = ["conv_nd", "max_pool", "batchnorm"]


try:
    from .gru import gru

    __all__ += ["gru"]
except ImportError:  # pragma: no cover
    pass



import numpy as np

from mygrad.nnet.initializers.uniform import uniform


def he_uniform(*shape, gain=1, dtype=np.float32, constant=None):
    r"""Initialize a ``mygrad.Tensor`` according to the uniform initialization procedure
    described by He et al.

    Parameters
    ----------
    shape : Sequence[int]
        The shape of the output Tensor. Note that ``shape`` must be at least two-dimensional.

    gain : Real, optional (default=1)
        The gain (scaling factor) to apply.

    dtype : data-type, optional (default=float32)
        The data type of the output tensor; must be a floating-point type.

    constant : bool, optional (default=False)
        If `True`, the returned tensor is a constant (it
            does not back-propagate a gradient).

    Returns
    -------
    mygrad.Tensor, shape=``shape``
        A Tensor, with values initialized according to the He uniform initialization.

    Notes
    -----
    He, Zhang, Ren, and Sun put forward this initialization in the paper
        "Delving Deep into Rectifiers: Surpassing Human-Level Performance
        on ImageNet Classification"
    https://arxiv.org/abs/1502.01852

    A Tensor :math:`W` initialized in this way should be drawn from a distribution about

    .. math::
        U[-\sqrt{\frac{6}{(1+a^2)n_l}}, \sqrt{\frac{6}{(1+a^2)n_l}}]

    where :math:`a` is the slope of the rectifier following this layer, which is incorporated
    using the `gain` variable above.

    The guidance put forward in that paper is that this initialization procedure should be prefered
    over the ``mygrad.nnet.initializers.glorot_*`` functions especially when rectifiers (e.g. ReLU,
    PReLU, leaky_relu) in very deep (> 1-20 or so layer) networks.

    Examples
    --------
    >>> from mygrad.nnet.initializers import he_uniform
    >>> he_uniform(2, 3)
    Tensor([[-0.97671795,  0.85518736, -0.8187388 ],
            [ 0.7599437 ,  0.94951814, -0.96755147]], dtype=float32)

    >>> he_uniform(4, 2, gain=5/3, dtype="float64", constant=True)
    Tensor([[-1.10372799, -0.16472136],
            [-1.32614867,  1.14142637],
            [ 0.78044471,  0.20562334],
            [-1.23968259,  1.0057054 ]])

    >>> he_uniform(2, 1, 2, dtype="float16")
    Tensor([[[-0.1233,  0.1023]],
            [[ 0.3845,  0.1003]]], dtype=float16)
    """
    if len(shape) == 1:
        shape = shape[0]
    if len(shape) < 2:
        raise ValueError("He Uniform initialization requires at least two dimensions")

    bound = gain / np.sqrt(3 / shape[1] * (np.prod(shape[2:]) if len(shape) > 2 else 1))
    return uniform(
        shape, lower_bound=-bound, upper_bound=bound, dtype=dtype, constant=constant
    )
